from ..receive import IReceiveListener, IReceiveService
from ..send import ISendService
from ..util.ConnectionDetails import ConnectionDetails


class LatencyClientService(IReceiveListener):
    def __init__(
            self,
            receive_service: IReceiveService,
            send_service: ISendService,
    ) -> None:
        self.send_service = send_service
        receive_service.add_receive_listener(self)

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.send_service.send(message)



from __future__ import annotations
from .interfaces import IGStreamerClientService, IGStreamerVideoListener
from .GStreamerRequestMessage import GStreamerRequestMessage
from .GStreamerResponseMessage import GstreamerResponseMessage
from .VideoSettings import VideoSettings
from ..constants import default_gstreamer_port
from ..receive import IReceiveListener, IReceiveService
from ..send import ISendService
from ..util.argument import AnyArguments, create_value_argument, IArguments
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Lazy import Lazy
from ..util.Listeners import Listeners


class GStreamerClientArguments(IArguments):
    def __init__(self) -> None:
        self.bit_rate = Lazy(lambda store: 2000000)
        self.fps = Lazy(lambda store: 30)
        self.height = Lazy(lambda store: 720)
        self.port = Lazy(lambda store: default_gstreamer_port)
        self.width = Lazy(lambda store: 1280)
        self.settings = Lazy(lambda store: VideoSettings(
            self.bit_rate(store),
            self.fps(store),
            self.height(store),
            self.width(store),
        ))

    def get_arguments(self) -> AnyArguments:
        return [
            create_value_argument(self.bit_rate, '--video-bit-rate', int, 'Video quality parameter.'),
            create_value_argument(self.fps, '--video-fps', int, 'Video frames per second.'),
            create_value_argument(self.height, '--video-height', int, 'Video resolution height.'),
            create_value_argument(self.port, '--video-port', int, 'Video stream port.'),
            create_value_argument(self.width, '--video-width', int, 'Video resolution width.'),
        ]


class GStreamerClientService(IGStreamerClientService, IReceiveListener):
    def __init__(
            self,
            arguments: GStreamerClientArguments,
            receive_service: IReceiveService,
            send_service: ISendService,
    ) -> None:
        self.arguments = arguments
        self.send_service = send_service
        self.listeners: Listeners[IGStreamerVideoListener] = Listeners()
        receive_service.add_receive_listener(self)

    def add_gstreamer_video_listener(self, listener: IGStreamerVideoListener) -> GStreamerClientService:
        self.listeners.add_listener(listener)
        return self

    def open_video(self, ip: str) -> None:
        self.send_service.send(GStreamerRequestMessage(
            (ip, self.arguments.port.get()), self.arguments.settings.get()
        ).encode())

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        response = GstreamerResponseMessage.decode(message)
        self.listeners.for_each(lambda listener: listener.on_video_available(response.caps, response.port))



from socket import socket
from typing import Callable, ContextManager
from .ServerArguments import ServerArguments
from ..car.CarServerService import CarServerService
from ..clock import IClock
from ..discovery.DiscoveryServerService import DiscoveryServerArguments, DiscoveryServerService
from ..expire.ExpireReceiveService import ExpireReceiveService
from ..expire.interfaces import IExpireReceiveService
from ..gpio.GpioServerService import GpioServerArguments, GpioServerService
from ..gpio.interfaces import IGpio, IGpioService
from ..gpio.PiGpio import PiGpio
from ..gstreamer.GStreamerServerService import GStreamerServerArguments, GStreamerServerService
from ..latency.interfaces import ILatencyServerService
from ..latency.LatencyServerService import LatencyServerArguments, LatencyServerService
from ..log.LogArguments import LogArguments
from ..message import message_types
from ..priority.PriorityReceiveService import PriorityReceiveService
from ..process import ProcessFactory
from ..receive import IReceiveService
from ..reliable import IReliableReceiveSendService
from ..routed.interfaces import IRoutedReceiveService, IRoutedSendService
from ..routed.RoutedSendService import RoutedSendService
from ..routed.RoutedReceiveService import RoutedReceiveService
from ..send import ISendService
from ..service import IServiceManager
from ..socket_ import ISocketFactory
from ..stop.interfaces import IStopServerService
from ..stop.StopServerService import StopServerService
from ..tcp.TcpServerService import TcpServerService
from ..throttle.ThrottleServerService import ThrottleServerArguments, ThrottleServerService
from ..timeout.interfaces import ITimeoutReceiveService
from ..timeout.TimeoutReceiveService import TimeoutReceiveService
from ..udp.UdpService import UdpService
from ..unreliable import IUnreliableReceiveSendService
from ..util.Lazy import Lazy
from ..util.MultiServiceManager import MultiServiceManager
from ..util.Placeholder import Placeholder
from ..util.Process import Process
from ..util.RTC import rtc
from ..util.Socket import SocketFactory


class Server:
    def __init__(self) -> None:
        self.discovery_arguments = DiscoveryServerArguments()
        self.gpio_arguments = GpioServerArguments()
        self.gstreamer_arguments = GStreamerServerArguments()
        self.latency_arguments = LatencyServerArguments()
        self.log_arguments = LogArguments()
        self.server_arguments = ServerArguments()
        self.throttle_arguments = ThrottleServerArguments()

        self.clock: Lazy[IClock] = Lazy(lambda store: rtc)
        self.process_factory: Lazy[ProcessFactory] = Lazy(lambda store: Process)
        self.socket_factory: Lazy[ISocketFactory] = Lazy(lambda store: SocketFactory(socket))
        self.gpio_factory: Lazy[Callable[[], IGpio]] = Lazy(lambda store: PiGpio)

        self.multi_service_manager = Lazy(lambda store: MultiServiceManager())
        self.service_manager: Lazy[IServiceManager] = Lazy(lambda store: self.multi_service_manager(store))

        self.discovery_service = Lazy(lambda store: DiscoveryServerService(
            self.discovery_arguments, self.server_arguments.listen_address(store)[1],
            self.socket_factory(store), self.service_manager(store)))

        self.gpio_service_implementation = Lazy(lambda store: GpioServerService(
            self.gpio_arguments, self.gpio_factory(store), self.service_manager(store)))
        self.gpio_service: Lazy[IGpioService] = Lazy(lambda store: self.gpio_service_implementation(store))

        self.tcp_service = Lazy(lambda store: TcpServerService(
            self.server_arguments.listen_address(store), self.service_manager(store), self.socket_factory(store)))
        self.reliable_service: Lazy[IReliableReceiveSendService] = Lazy(lambda store: self.tcp_service(store))

        self.udp_service = Lazy(lambda store: UdpService(
            self.clock(store), True, Placeholder(self.server_arguments.listen_address(store)),
            self.service_manager(store), self.socket_factory(store),
            self.server_arguments.unreliable_timeout_seconds(store)))
        self.unreliable_service: Lazy[IUnreliableReceiveSendService] = Lazy(lambda store: self.udp_service(store))

        self.reliable_routed_receive_service: Lazy[IRoutedReceiveService] = Lazy(lambda store: RoutedReceiveService(
            self.reliable_service(store)))
        self.reliable_routed_send_service: Lazy[IRoutedSendService] = Lazy(lambda store: RoutedSendService(
            self.reliable_service(store)))
        self.unreliable_routed_receive_service: Lazy[IRoutedReceiveService] = Lazy(lambda store: RoutedReceiveService(
            self.unreliable_service(store)))
        self.unreliable_routed_send_service: Lazy[IRoutedSendService] = Lazy(lambda store: RoutedSendService(
            self.unreliable_service(store)))

        self.latency_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.unreliable_routed_receive_service(store).create_receive_service(message_types.latency))
        self.latency_send_service: Lazy[ISendService] = Lazy(
            lambda store: self.unreliable_routed_send_service(store).create_send_service(message_types.latency))
        self.latency_service_implementation = Lazy(lambda store: LatencyServerService(
            self.latency_arguments, self.clock(store), self.latency_receive_service(store),
            self.latency_send_service(store), self.service_manager(store)))
        self.latency_service: Lazy[ILatencyServerService] = Lazy(
            lambda store: self.latency_service_implementation(store))

        self.car_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.unreliable_routed_receive_service(store).create_receive_service(message_types.car))

        self.priority_service_implementation = Lazy(lambda store: PriorityReceiveService(
            self.clock(store), self.car_receive_service(store), self.service_manager(store),
            self.server_arguments.priority_timeout_seconds(store)))
        self.priority_service: Lazy[IReceiveService] = Lazy(lambda store: self.priority_service_implementation(store))

        self.expire_service_implementation = Lazy(lambda store: ExpireReceiveService(self.priority_service(store)))
        self.expire_service: Lazy[IExpireReceiveService] = Lazy(lambda store: self.expire_service_implementation(store))

        self.timeout_service_implementation = Lazy(lambda store: TimeoutReceiveService(
            self.clock(store), self.expire_service(store), self.service_manager(store)))
        self.timeout_service: Lazy[ITimeoutReceiveService] = Lazy(
            lambda store: self.timeout_service_implementation(store))

        self.throttle_service_implementation = Lazy(lambda store: ThrottleServerService(
            self.throttle_arguments, self.gpio_service(store), self.latency_service(store),
            self.timeout_service(store)))
        self.throttle_service: Lazy[IGpioService] = Lazy(lambda store: self.throttle_service_implementation(store))

        self.car_service = Lazy(lambda store: CarServerService(
            self.expire_service(store), self.throttle_service(store), self.reliable_service(store),
            self.timeout_service(store)))

        self.gstreamer_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.reliable_routed_receive_service(store).create_receive_service(message_types.gstreamer))
        self.gstreamer_send_service: Lazy[ISendService] = Lazy(
            lambda store: self.reliable_routed_send_service(store).create_send_service(message_types.gstreamer))
        self.gstreamer_service = Lazy(lambda store: GStreamerServerService(
            self.gstreamer_arguments, self.process_factory(store), self.gstreamer_receive_service(store),
            self.reliable_service(store), self.gstreamer_send_service(store), self.service_manager(store)))

        self.stop_receive_service: Lazy[IReceiveService] = Lazy(
            lambda store: self.reliable_routed_receive_service(store).create_receive_service(message_types.stop))
        self.stop_service_implementation = Lazy(lambda store: StopServerService(
            self.stop_receive_service(store), self.service_manager(store)))
        self.stop_service: Lazy[IStopServerService] = Lazy(lambda store: self.stop_service_implementation(store))

    def use_services(self) -> ContextManager[None]:
        return self.clock.get().use_services(self.service_manager.get())



from __future__ import annotations
from abc import ABC, abstractmethod
from ..receive import IReceiveService
from ..send import ISendService
from ..util.ConnectionDetails import ConnectionDetails


class IRoutedReceiveListener(ABC):
    @abstractmethod
    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        """"""


class IRoutedReceiveService(ABC):
    @abstractmethod
    def add_receive_listener(self, message_type: int, listener: IRoutedReceiveListener) -> IRoutedReceiveService:
        """"""

    @abstractmethod
    def create_receive_service(self, message_type: int) -> IReceiveService:
        """"""


class IRoutedSendService(ABC):
    @abstractmethod
    def create_send_service(self, message_type: int) -> ISendService:
        """"""

    @abstractmethod
    def send(self, message_type: int, message: bytes) -> None:
        """"""



from __future__ import annotations
from abc import ABC, abstractmethod


class IGStreamerVideoListener(ABC):
    @abstractmethod
    def on_video_available(self, caps: str, port: int) -> None:
        """"""


class IGStreamerClientService(ABC):
    @abstractmethod
    def add_gstreamer_video_listener(self, listener: IGStreamerVideoListener) -> IGStreamerClientService:
        """"""

    @abstractmethod
    def open_video(self, ip: str) -> None:
        """"""



from __future__ import annotations
from typing import Callable, Optional
from .interfaces import IGpio, IGpioService
from ..service import IService, IServiceManager
from ..util.argument import IArguments
from ..util.Placeholder import Placeholder
from ..util.util import min_max
from ..util.Lazy import Lazy


class GpioServerArguments(IArguments):
    def __init__(self) -> None:
        self.pwm_base = Lazy(lambda store: 300)
        self.pwm_frequency = Lazy(lambda store: 50)
        self.pwm_range = Lazy(lambda store: 4000)
        self.pwm_motor_base = Lazy(lambda store: self.pwm_base(store))
        self.pwm_motor_frequency = Lazy(lambda store: self.pwm_frequency(store))
        self.pwm_motor_gpio = Lazy(lambda store: 12)
        self.pwm_motor_invert = Lazy(lambda store: False)
        self.pwm_motor_maximum = Lazy(lambda store: 100)
        self.pwm_motor_minimum = Lazy(lambda store: -self.pwm_motor_maximum(store))
        self.pwm_motor_offset = Lazy(lambda store: 0)
        self.pwm_motor_range = Lazy(lambda store: self.pwm_range(store))
        self.pwm_steering_base = Lazy(lambda store: self.pwm_base(store))
        self.pwm_steering_frequency = Lazy(lambda store: self.pwm_frequency(store))
        self.pwm_steering_gpio = Lazy(lambda store: 13)
        self.pwm_steering_invert = Lazy(lambda store: False)
        self.pwm_steering_maximum = Lazy(lambda store: 100)
        self.pwm_steering_minimum = Lazy(lambda store: -self.pwm_steering_maximum(store))
        self.pwm_steering_offset = Lazy(lambda store: 0)
        self.pwm_steering_range = Lazy(lambda store: self.pwm_range(store))


class GpioServerService(IGpioService, IService):
    def __init__(
            self, arguments: GpioServerArguments, gpio_factory: Callable[[], IGpio],
            service_manager: IServiceManager,
    ) -> None:
        self.arguments = arguments
        self.gpio_factory = gpio_factory
        self.gpio: Placeholder[IGpio] = Placeholder()
        service_manager.add_service(self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return self.gpio.is_present()

    def start_service(self) -> None:
        gpio = self.gpio_factory()
        gpio.set_mode(self.arguments.pwm_motor_gpio.get(), 1)
        gpio.set_pwm_frequency(self.arguments.pwm_motor_gpio.get(), self.arguments.pwm_motor_frequency.get())
        gpio.set_pwm_range(self.arguments.pwm_motor_gpio.get(), self.arguments.pwm_motor_range.get())
        _set_motor(self.arguments, gpio, 0)
        gpio.set_mode(self.arguments.pwm_steering_gpio.get(), 1)
        gpio.set_pwm_frequency(self.arguments.pwm_steering_gpio.get(), self.arguments.pwm_steering_frequency.get())
        gpio.set_pwm_range(self.arguments.pwm_steering_gpio.get(), self.arguments.pwm_steering_range.get())
        _set_steering(self.arguments, gpio, 0)
        self.gpio.set(gpio)

    def update(self, speed: int, steering: int) -> None:
        with self.gpio as (gpio, _):
            if gpio is not None:
                _set_motor(self.arguments, gpio, speed)
                _set_steering(self.arguments, gpio, steering)

    def stop_service(self) -> None:
        self.reset()
        self.gpio.get_and_clear().stop()


def _set_motor(arguments: GpioServerArguments, gpio: IGpio, value: int) -> None:
    factor = -1 if arguments.pwm_motor_invert.get() else 1
    pwm_value = arguments.pwm_motor_base.get() + arguments.pwm_motor_offset.get() + min_max(
        value * factor, arguments.pwm_motor_minimum.get(), arguments.pwm_motor_maximum.get())
    gpio.set_pwm_duty_cycle(arguments.pwm_motor_gpio.get(), pwm_value)


def _set_steering(arguments: GpioServerArguments, gpio: IGpio, value: int) -> None:
    factor = -1 if arguments.pwm_steering_invert.get() else 1
    pwm_value = arguments.pwm_steering_base.get() + arguments.pwm_steering_offset.get() + min_max(
        value * factor, arguments.pwm_steering_minimum.get(), arguments.pwm_steering_maximum.get())
    gpio.set_pwm_duty_cycle(arguments.pwm_steering_gpio.get(), pwm_value)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional


class IService(ABC):
    @abstractmethod
    def get_service_name(self) -> str:
        """"""

    @abstractmethod
    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        """
        :return: True if services is still running, False otherwise.
        """

    @abstractmethod
    def start_service(self) -> None:
        """"""

    @abstractmethod
    def stop_service(self) -> None:
        """"""


class IStartedService(ABC):
    @abstractmethod
    def get_service_name(self) -> str:
        """"""

    @abstractmethod
    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        """"""

    @abstractmethod
    def stop_service(self) -> None:
        """"""


class StartedServiceWrapper(IStartedService):
    def __init__(self, service: IService, service_manager: IServiceManager) -> None:
        self.service = service
        service_manager.add_started_service(self)

    def get_service_name(self) -> str:
        return self.service.get_service_name()

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        return self.service.join_service(timeout_seconds)

    def stop_service(self) -> None:
        self.service.stop_service()


class IServiceManager(ABC):
    @abstractmethod
    def add_service(self, service: IService) -> None:
        """"""

    @abstractmethod
    def add_started_service(self, service: IStartedService) -> None:
        """"""

    @abstractmethod
    def join_all(self) -> None:
        """"""

    @abstractmethod
    def start_all(self) -> None:
        """"""

    @abstractmethod
    def stop_all(self) -> None:
        """"""



from __future__ import annotations
from queue import Empty
from threading import Thread
from typing import Optional
from .interfaces import ILatencyListener
from .interfaces import ILatencyServerService
from .LatencyMessage import LatencyMessage
from ..clock import IClock
from ..queue_ import IQueue
from ..receive import IReceiveListener, IReceiveService
from ..send import ISendService
from ..service import IService, IServiceManager
from ..util.argument import IArguments
from ..util.ConnectionDetails import ConnectionDetails
from ..util.InterruptableSleep import InterruptableSleep
from ..util.Lazy import Lazy
from ..util.IdealQueue import IdealQueue
from ..util.Listeners import Listeners


class LatencyServerArguments(IArguments):
    def __init__(self) -> None:
        self.latency_check_interval_seconds = Lazy(lambda store: 1.0)


class LatencyServerService(ILatencyServerService, IReceiveListener, IService):
    def __init__(
            self,
            arguments: LatencyServerArguments,
            clock: IClock,
            receive_service: IReceiveService,
            send_service: ISendService,
            service_manager: IServiceManager,
    ) -> None:
        self.arguments = arguments
        self.clock = clock
        self.interruptable_sleep = InterruptableSleep(clock)
        self.listeners: Listeners[ILatencyListener] = Listeners()
        self.received_messages: IQueue[Optional[LatencyMessage]] = IdealQueue()
        self.send_service = send_service
        self.thread = Thread(target=self.run)
        receive_service.add_receive_listener(self)
        service_manager.add_service(self)

    def get_service_name(self) -> str:
        return __name__

    def join_service(self, timeout_seconds: Optional[float] = None) -> bool:
        self.thread.join(timeout_seconds)
        return self.thread.is_alive()

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        self.received_messages.put(LatencyMessage.decode(message))

    def run(self) -> None:
        sent_message = LatencyMessage(0)
        while True:
            try:
                sent_message.increment()
                sent_timestamp = self.clock.get_seconds()
                self.send_service.send(sent_message.encode())
                while True:
                    timeout_seconds = sent_timestamp + self.arguments.latency_check_interval_seconds.get(
                    ) - self.clock.get_seconds()
                    if timeout_seconds <= 0:
                        raise Empty()
                    received_message = self.clock.get_from_queue(self.received_messages, timeout_seconds)
                    received_timestamp = self.clock.get_seconds()
                    if received_message is None:
                        return
                    if received_message.number == sent_message.number:
                        break
                self.listeners.for_each(lambda listener: listener.on_latency_available(
                    received_timestamp - sent_timestamp))
                if not self.interruptable_sleep.sleep(self.arguments.latency_check_interval_seconds.get()):
                    return
            except Empty:
                self.listeners.for_each(lambda listener: listener.on_latency_timeout())

    def start_service(self) -> None:
        self.thread.start()

    def stop_service(self) -> None:
        self.interruptable_sleep.interrupt()
        self.received_messages.put(None)

    def add_latency_listener(self, latency_listener: ILatencyListener) -> LatencyServerService:
        self.listeners.add_listener(latency_listener)
        return self



from __future__ import annotations
from typing import Dict
from .interfaces import IRoutedReceiveListener, IRoutedReceiveService
from .RoutedMessage import RoutedMessage
from ..receive import IReceiveListener, IReceiveService
from ..util.ConnectionDetails import ConnectionDetails
from ..util.Listeners import Listeners


class ReceiveListener(IRoutedReceiveListener):
    def __init__(self, listener: IReceiveListener) -> None:
        self.listener = listener

    def on_routed_receive(self, message_type: int, message: bytes, details: ConnectionDetails) -> None:
        self.listener.on_receive(message, details)


class ReceiveService(IReceiveService):
    def __init__(self, message_type: int, routed_receive_service: RoutedReceiveService) -> None:
        self.message_type = message_type
        self.routed_receive_service = routed_receive_service

    def add_receive_listener(self, listener: IReceiveListener) -> ReceiveService:
        self.routed_receive_service.add_receive_listener(self.message_type, ReceiveListener(listener))
        return self


class RoutedReceiveService(IReceiveListener, IRoutedReceiveService):
    def __init__(self, receive_service: IReceiveService) -> None:
        self.received_listeners: Dict[int, Listeners[IRoutedReceiveListener]] = dict()
        receive_service.add_receive_listener(self)

    def add_receive_listener(self, message_type: int, listener: IRoutedReceiveListener) -> RoutedReceiveService:
        if message_type > 255:
            raise RuntimeError(f'Illegal message type: {message_type} > 255.')
        if message_type not in self.received_listeners:
            self.received_listeners[message_type] = Listeners()
        self.received_listeners[message_type].add_listener(listener)
        return self

    def create_receive_service(self, message_type: int) -> IReceiveService:
        return ReceiveService(message_type, self)

    def on_receive(self, message: bytes, details: ConnectionDetails) -> None:
        routed_message = RoutedMessage.decode(message)
        self.received_listeners[routed_message.message_type].for_each(
            lambda listener: listener.on_routed_receive(routed_message.message_type, routed_message.payload, details))



from __future__ import annotations
from logging import getLogger
from .CarMessage import CarMessage
from ..gpio.interfaces import IGpioService
from ..reliable import IReliableConnectListener, IReliableDisconnectListener, IReliableService
from ..timeout.interfaces import ITimeoutSendService
from ..util.ConnectionDetails import ConnectionDetails


class CarClientService(IGpioService, IReliableConnectListener, IReliableDisconnectListener):
    def __init__(
            self,
            reliable_service: IReliableService,
            timeout_send_service: ITimeoutSendService,
    ) -> None:
        self._car_message = CarMessage(0, 0)
        self.logger = getLogger(__name__)
        self.timeout_send_service = timeout_send_service
        reliable_service.add_reliable_connect_listener(self).add_reliable_disconnect_listener(self)

    def on_reliable_connect(self, details: ConnectionDetails) -> None:
        self.timeout_send_service.set_and_send_immediately(lambda: self._car_message.encode())

    def on_reliable_disconnect(self, details: ConnectionDetails) -> None:
        self.timeout_send_service.set_and_send_immediately(lambda: None)
        self.logger.info(f'Client statistics: sent={self.timeout_send_service.get_send_count()}')

    def update(self, speed: int, steering: int) -> None:
        self.timeout_send_service.set_and_send_immediately(lambda: self._car_message.update(speed, steering).encode())



from ..constants import default_server_port, discover_server_ip
from ..util.argument import AnyArguments, create_value_argument, IArguments
from ..util.Lazy import Lazy


class ClientArguments(IArguments):
    def __init__(self) -> None:
        self.priority = Lazy(lambda store: 128)
        self.reconnect_seconds = Lazy(lambda store: 1.0)
        self.send_timeout_seconds = Lazy(lambda store: 0.125)
        self.server_ip = Lazy(lambda store: discover_server_ip)
        self.server_port = Lazy(lambda store: default_server_port)
        self.server_address = Lazy(lambda store: (self.server_ip(store), self.server_port(store)))

    def get_arguments(self) -> AnyArguments:
        return [
            create_value_argument(self.priority, '--priority', int, 'Priority of client.'),
            create_value_argument(self.server_ip, '--server-ip', str, 'TCP/UDP server IP.'),
            create_value_argument(self.server_port, '--server-port', int, 'TCP/UDP server port.'),
        ]
